/**
  *  \file server/server.cc
  *  \brief PCC2 Server
  *
  *  This opens a TRN/RST pair and talks stdin/stdout.
  */

#include <cstring>
#include <iostream>
#include <cstdlib>
#include "arch/platform.h"
#include "util/misc.h"
#include "util/console.h"
#include "util/modinit.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "util/recode.h"
#include "util/except.h"
#include "game/global.h"
#include "game/load.h"
#include "game/gamedirs.h"
#include "game/storage.h"
#include "game/team.h"
#include "game/db.h"
#include "game/config.h"
#include "game/mission.h"
#include "game/msgglobal.h"
#include "game/racename.h"
#include "game/reg.h"
#include "game/sel.h"
#include "game/stat.h"
#include "int/intglobal.h"
#include "server/serverconsole.h"
#include "server/serverio.h"
#include "server/getobj.h"
#include "server/jsonin.h"
#include "server/postobj.h"
#include "server/getquery.h"
#include "game/fcode.h"
#include "game/storage/overview.h"
#include "server/prop.h"

// Constraints: not with mod_base_gui, not with mod_base_console
static const ModuleInitializer mod_base_server[] = {
    { OnStartup, Platform::initPlatform },
    { OnStartup, initServerConsole },
    { OnStartup, initNLS },
    { 0, 0 },
};

static const char* arg_progname;

static void
die(const string_t s)
{
    setServerConsoleState(true);
    console.write(LOG_ERROR, format("%s: %s", arg_progname, s));
    std::exit(2);
}

static void
help()
{
    console << format(_("PCC2 Server %s - (c) 2011 Stefan Reuther\n\n"
                        "Usage:\n"
                        "  %s [-CCHARSET] PLAYER GAMEDIR [ROOTDIR]\n\n"
                        "Report bugs to <Streu@gmx.de>\n"),
                      arg_progname);
}

static void
doHelp()
{
    writeClient("200 OK, help follows");
    writeClient("  HELP   this message\n"
                "  SAVE   save data\n"
                "  STAT   status (show console)\n"
                "  GET x  data access\n"
                "     obj/id,id,id   dynamic JSON data\n"
                "     item/id        rendered static data\n"
                "     query/q        query\n"
                "  POST x data modify (content follows)\n"
                "     obj/id         modify object");
    writeClientTerminator();
}

static void
performPostTest(const string_t& body)
{
    writeClient("200 Test output follows");
    JsonTokenizer tok(body);
    while (tok.readNext() != tok.kEOF) {
        switch (tok.getCurrent()) {
         case JsonTokenizer::kError:
            writeClient("error"); break;
         case JsonTokenizer::kEOF:
            writeClient("eof"); break;
         case JsonTokenizer::kString:
            writeClient(format("string '%s'", tok.getStringValue())); break;
         case JsonTokenizer::kKeyword:
            writeClient(format("keyword '%s'", tok.getStringValue())); break;
         case JsonTokenizer::kInt:
            writeClient(format("int %d", tok.getIntValue())); break;
         case JsonTokenizer::kFloat:
            writeClient(format("float %.4f", tok.getFloatValue())); break;
         case JsonTokenizer::kStartArray:
            writeClient("startArray"); break;
         case JsonTokenizer::kEndArray:
            writeClient("endArray"); break;
         case JsonTokenizer::kStartHash:
            writeClient("startHash"); break;
         case JsonTokenizer::kEndHash:
            writeClient("endHash"); break;
         case JsonTokenizer::kComma:
            writeClient("comma"); break;
         case JsonTokenizer::kColon:
            writeClient("colon"); break;
         default:
            writeClient("huh?"); break;
        }
    }
    writeClientTerminator();
}

static void
doSave()
{
    GGameDataStorage::instance->save(getCurrentTurn());
    writeClient("100 OK, data saved");
}

static void
doStat()
{
    writeClient("200 OK, console follows");
    setServerConsoleState(true);
    writeClient(format("Bytes transferred so far: read %d, written %d", bytesRead, bytesWritten));
    setServerConsoleState(false);
    writeClientTerminator();
}

static void
doGet(const string_t& args)
{
    if (args.size() > 4 && args.compare(0, 4, "obj/", 4) == 0) {
        performGetObject(args.substr(4));
    } else if (args.size() > 5 && args.compare(0, 5, "item/", 5) == 0) {
        writeClient("501 Not implemented");
    } else if (args.size() > 6 && args.compare(0, 6, "query/", 6) == 0) {
        performGetQuery(args.substr(6));
    } else {
        writeClient("404 Not found");
    }
}

static void
doPost(const string_t& args)
{
    /* Read message body */
    /* FIXME: limit size */
    string_t body;
    string_t line;
    while (readClient(line) && line != ".") {
        body += line;
        body += '\n';
    }

    /* Dispatch */
    if (ServerObjectCommandHandler* och = createCommandHandler(args)) {
        doPostObject(och, body);
    } else if (args == "test") {
        performPostTest(body);
    } else {
        writeClient("404 Not found");
    }
}

static void
play()
{
    /* Server main routine */
    string_t command;
    writeClient("100 OK");
    while (readClient(command)) {
        string_t args;
        strSplit(command, command, args, " ");
        if (strCaseCmp(command, "HELP") == 0) {
            doHelp();
        } else if (strCaseCmp(command, "SAVE") == 0) {
            doSave();
        } else if (strCaseCmp(command, "STAT") == 0) {
            doStat();
        } else if (strCaseCmp(command, "GET") == 0) {
            doGet(args);
        } else if (strCaseCmp(command, "POST") == 0) {
            doPost(args);
        } else if (strCaseCmp(command, "QUIT") == 0) {
            break;
        } else {
            writeClient("400 Bad Request");
        }
    }
}

static const ModuleInitializer* server_modules[] = {
    mod_base_server,
    mod_config,
    mod_coord,
    mod_int,
    mod_missions,
    mod_msg,
    mod_msgglobal,
    mod_racenames,
    mod_reg,
    mod_sel,
    mod_spec,
    mod_stat,
    mod_team,
    mod_fcode,
    mod_prop,
    0
};

int main(int argc, char** argv)
{
    /* Initialize */
    performInitialisation(OnStartup, server_modules);
    setServerConsoleState(true);

    /* Parse arguments */
    argv = Platform::convertArgumentsToUtf8(argv);
    const char* arg_dir = 0;
    const char* arg_root = 0;
    int arg_player = 0;
    int n;
    arg_progname = argv[0];
    while (const char* p = *++argv) {
        if (p[0] == '-' && p[1] == '-')
            ++p;
        if (std::strcmp(p, "-h") == 0 || std::strcmp(p, "-help") == 0) {
            help();
            return 0;
        } else if (strncmp(p, "-C", 2) == 0) {
            const char* cs = p+2;
            if (!*p) {
                cs = *++argv;
                if (!cs)
                    die(_("option `-C' needs an argument (the character set)"));
            }
            CharacterSet cset = CharacterSet::getCharsetByName(cs);
            if (!cset.isValid()) {
                die(_("the specified character set is not known"));
            }
            setGameCharacterSet(cset);
        } else if (strncmp(p, "-R", 2) == 0 || strncmp(p, "-W", 2) == 0) {
            /* session conflict management; skip those */
        } else if (*p == '-') {
            die(format(_("invalid option specified. Use `%s -h' for help"), arg_progname));
        } else if (stringValue(p, n) && n > 0 && n <= 11) {
            if (arg_player != 0) {
                die(_("only one player number allowed"));
            }
            arg_player = n;
        } else if (!arg_dir) {
            arg_dir = p;
        } else if (!arg_root) {
            arg_root = p;
        } else {
            die(_("too many arguments"));
        }
    }

    if (!arg_player)
        die(_("missing player number"));
    if (!arg_dir)
        die(_("missing directory name"));

    /* Initialize and load game */
    setServerConsoleState(false);
    try {
        /* Scan directory. Not very efficient, but gets the job done. */
        initGameGlobals(arg_dir, arg_root ? string_t(arg_root) : Platform::getDataDirectory());
        GDirectoryOverview ov;
        ov.scan(*game_file_dir, false);
        if ((ov.getPlayerFlags(arg_player) & ov.have_RST) == 0) {
            die(_("no result file for this player"));
        }

        /* Setup load */
        setPlayerIds(arg_player, arg_player);
        performInitialisation(OnEnterDirectory | OnEnterGame, server_modules);
        GGameDataStorage::instance =
            new GGameResultStorage(arg_player,
                                   (ov.getPlayerFlags(arg_player) & ov.have_TRN));

        /* Load (copied from main.cc) */
        loadCommon(getCurrentTurn(), *game_file_dir, *spec_file_dir, LoadBoth, arg_player);
        if (Ptr<Stream> cdb = game_file_dir->openFileNT(format("chart%d.cc", arg_player), Stream::C_READ))
            loadChartDatabase(*cdb, getCurrentTurn(), true);
        GGameDataStorage::instance->load(getCurrentTurn());
        getCurrentTurn().postprocess(GObject::Playable);

        /* Load properties */
        if (Ptr<Stream> prop = game_file_dir->openFileNT(".c2file", Stream::C_READ)) {
            loadProperties(*prop);
        }

        /* Display */
        setDisplayedTurn(0);
        performInitialisation(OnEnteredGame, server_modules);
    }
    catch (FileProblemException& e) {
        die(format(_("error while accessing file %s: %s"), e.getFileName(), e.what()));
    }
    catch (std::exception& e) {
        die(format(_("error during intialisation: %s"), e.what()));
    }

    /* Play! */
    play();

    /* Close */
    try {
        performInitialisation(OnLeaveGame, server_modules);
        GGameDataStorage::instance->save(getCurrentTurn());
        if (Ptr<Stream> cdb = game_file_dir->openFileNT(format("chart%d.cc", arg_player),
                                                        Stream::C_CREATE))
            saveChartDatabase(*cdb, getCurrentTurn());
        performInitialisation(OnLeftGame | OnLeftDirectory | OnShutdown, server_modules);
        GGameDataStorage::instance = 0;
    }
    catch (FileProblemException& e) {
        die(format(_("error while accessing file %s: %s"), e.getFileName(), e.what()));
    }
    catch (std::exception& e) {
        die(format(_("error during shutdown: %s"), e.what()));
    }

    return 0;
}
